# -*- coding: utf-8 -*-
# Author: Nianze A. Tao (Omozawa Sueno)
"""
Define ChemBFN and regressor models for training.
"""
from pathlib import Path
from typing import Dict, Tuple, Union, Optional
import torch
import torch.optim as op
import torch.nn.functional as F
from loralib import lora_state_dict, mark_only_lora_as_trainable
from torch import Tensor
from torch.optim.lr_scheduler import ReduceLROnPlateau
from lightning import LightningModule
from .model import ChemBFN, MLP
from .scorer import Scorer

DEFAULT_MODEL_HPARAM = {"lr": 5e-5, "lr_warmup_step": 1000, "uncond_prob": 0.2}
DEFAULT_REGRESSOR_HPARAM = {
    "mode": "regression",
    "lr_scheduler_factor": 0.8,
    "lr_scheduler_patience": 20,
    "lr_warmup_step": 1000,
    "max_lr": 1e-4,
    "freeze": False,
}


class Model(LightningModule):
    def __init__(
        self,
        model: ChemBFN,
        mlp: Optional[MLP] = None,
        scorer: Optional[Scorer] = None,
        hparam: Dict[str, Union[int, float]] = DEFAULT_MODEL_HPARAM,
    ) -> None:
        """
        A `~lightning.LightningModule` wrapper of bayesian flow network for chemistry model.\n
        This module is used in training stage only. By calling `Model(...).export_model(YOUR_WORK_DIR)` after training,
        the model(s) will be saved to `YOUR_WORK_DIR/model.pt` (if LoRA is enabled then `YOUR_WORK_DIR/lora.pt`)
        and (if exists) `YOUR_WORK_DIR/mlp.pt`.

        :param model: `~bayesianflow_for_chem.model.ChemBFN` instance.
        :param mlp: `~bayesianflow_for_chem.model.MLP` instance or `None`.
        :param scorer: `~bayesianflow_for_chem.scorer.Scorer` instance or `None`.
        :param hparam: a `dict` instance of hyperparameters. See `bayesianflow_for_chem.train.DEFAULT_MODEL_HPARAM`.
        :type model: bayesianflow_for_chem.model.ChemBFN
        :type mlp: bayesianflow_for_chem.model.MLP | None
        :type scorer: bayesianflow_for_chem.scorer.Scorer | None
        :type hparam: dict
        """
        super().__init__()
        self.model = model
        self.mlp = mlp
        self.scorer = scorer
        self.save_hyperparameters(hparam, ignore=["model", "mlp", "scorer"])
        if model.lora_enabled:
            mark_only_lora_as_trainable(self.model)
        self.use_scorer = self.scorer is not None

    def training_step(self, batch: Dict[str, Tensor]) -> Tensor:
        x = batch["token"]
        t = torch.rand((x.shape[0], 1, 1), device=x.device)
        if "mask" in batch:
            mask = batch["mask"]
        else:
            mask = None
        if self.mlp is not None:
            y = batch["value"]
            y = self.mlp.forward(y)
            if y.dim() == 2:
                y = y[:, None, :]
            y_mask = F.dropout(torch.ones_like(t), self.hparams.uncond_prob, True, True)
            y_mask = (y_mask != 0).float()
            loss, p = self.model.cts_loss(x, t, y * y_mask, mask, self.use_scorer)
        else:
            loss, p = self.model.cts_loss(x, t, None, mask, self.use_scorer)
        self.log("continuous_time_loss", loss.item())
        if self.use_scorer:
            scorer_loss = self.scorer.calc_score_loss(p)
            self.log(f"{self.scorer.name}_loss", scorer_loss.item())
            loss += scorer_loss * self.scorer.eta
        return loss

    def configure_optimizers(self) -> Dict[str, op.AdamW]:
        optimizer = op.AdamW(self.parameters(), lr=1e-8, weight_decay=0.01)
        return {"optimizer": optimizer}

    def optimizer_step(self, *args, **kwargs) -> None:
        optimizer: op.AdamW = kwargs["optimizer"] if "optimizer" in kwargs else args[2]
        # warm-up step
        if self.trainer.global_step < self.hparams.lr_warmup_step:
            lr_scale = int(self.trainer.global_step + 1) / self.hparams.lr_warmup_step
            lr_scale = min(1.0, lr_scale)
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.lr
        super().optimizer_step(*args, **kwargs)
        optimizer.zero_grad(set_to_none=True)

    def export_model(self, workdir: Path) -> None:
        """
        Save the trained model.

        :param workdir: the directory to save the model(s)
        :type workdir: pathlib.Path
        :return:
        :rtype: None
        """
        if self.model.lora_enabled:
            torch.save(
                {
                    "lora_nn": lora_state_dict(self.model),
                    "lora_param": self.model.lora_param,
                },
                workdir / "lora.pt",
            )
        else:
            torch.save(
                {"nn": self.model.state_dict(), "hparam": self.model.hparam},
                workdir / "model.pt",
            )
        if self.mlp is not None:
            torch.save(
                {"nn": self.mlp.state_dict(), "hparam": self.mlp.hparam},
                workdir / "mlp.pt",
            )


class Regressor(LightningModule):
    def __init__(
        self,
        model: ChemBFN,
        mlp: MLP,
        hparam: Dict[str, Union[str, int, float, bool]] = DEFAULT_REGRESSOR_HPARAM,
    ) -> None:
        """
        A `~lightning.LightningModule` wrapper of bayesian flow network for chemistry regression model.\n
        This module is used in training stage only. By calling `Regressor(...).export_model(YOUR_WORK_DIR)` after training,
        the models will be saved to `YOUR_WORK_DIR/model_ft.pt` (if LoRA is enabled then `YOUR_WORK_DIR/lora.pt`)
        and `YOUR_WORK_DIR/readout.pt`.

        :param model: `~bayesianflow_for_chem.model.ChemBFN` instance.
        :param mlp: `~bayesianflow_for_chem.model.MLP` instance.
        :param hparam: a `dict` instance of hyperparameters. See `bayesianflow_for_chem.train.DEFAULT_REGRESSOR_HPARAM`.
        :type model: bayesianflow_for_chem.model.ChemBFN
        :type mlp: bayesianflow_for_chem.model.MLP
        :type hparam: dict
        """
        super().__init__()
        self.model = model
        self.mlp = mlp
        self.model.requires_grad_(not hparam["freeze"])
        self.save_hyperparameters(hparam, ignore=["model", "mlp"])
        if model.lora_enabled:
            mark_only_lora_as_trainable(self.model)
        assert hparam["mode"] in ("regression", "classification")

    @staticmethod
    def _mask_label(label: Tensor) -> Tuple[Tensor, Tensor]:
        # find the unlabelled position(s)
        label_mask = (label != torch.inf).float()
        # masked the unlabelled position(s)
        masked_label = label.masked_fill(label == torch.inf, 0)
        return label_mask, masked_label

    def training_step(self, batch: Dict[str, Tensor]) -> Tensor:
        x, y = batch["token"], batch["value"]
        z = self.model.inference(x, self.mlp)
        if self.hparams.mode == "classification":
            n_b, n_y = y.shape
            z = z.reshape(n_b * n_y, -1)
            loss = F.cross_entropy(z, y.reshape(-1).to(torch.long))
        else:
            y_mask, y = self._mask_label(y)
            loss = F.mse_loss(z * y_mask, y, reduction="mean")
        self.log("train_loss", loss.item())
        return loss

    def validation_step(self, batch: Dict[str, Tensor]) -> None:
        x, y = batch["token"], batch["value"]
        z = self.model.inference(x, self.mlp)
        if self.hparams.mode == "classification":
            n_b, n_y = y.shape
            z = z.reshape(n_b * n_y, -1)
            val_loss = 1 - (torch.argmax(z, -1) == y.reshape(-1)).float().mean()
        else:
            y_mask, y = self._mask_label(y)
            val_loss = (z * y_mask - y).abs().sum() / y_mask.sum()
        self.log("val_loss", val_loss.item())

    def configure_optimizers(self) -> Dict:
        optimizer = op.AdamW(self.parameters(), lr=1e-7, weight_decay=0.01)
        lr_scheduler_config = {
            "scheduler": ReduceLROnPlateau(
                optimizer,
                "min",
                factor=self.hparams.lr_scheduler_factor,
                patience=self.hparams.lr_scheduler_patience,
                min_lr=1e-6,
            ),
            "interval": "epoch",
            "monitor": "val_loss",
            "frequency": 1,
            "strict": True,
        }
        return {"optimizer": optimizer, "lr_scheduler": lr_scheduler_config}

    def optimizer_step(self, *args, **kwargs) -> None:
        optimizer: op.AdamW = kwargs["optimizer"] if "optimizer" in kwargs else args[2]
        # warm-up step
        if self.trainer.global_step < self.hparams.lr_warmup_step:
            lr_scale = int(self.trainer.global_step + 1) / self.hparams.lr_warmup_step
            lr_scale = min(1.0, lr_scale)
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.max_lr
        super().optimizer_step(*args, **kwargs)
        optimizer.zero_grad(set_to_none=True)

    def export_model(self, workdir: Path) -> None:
        """
        Save the trained model.

        :param workdir: the directory to save the models
        :type workdir: pathlib.Path
        :return:
        :rtype: None
        """
        torch.save(
            {"nn": self.mlp.state_dict(), "hparam": self.mlp.hparam},
            workdir / "readout.pt",
        )
        if not self.hparams.freeze:
            if self.model.lora_enabled:
                torch.save(
                    {
                        "lora_nn": lora_state_dict(self.model),
                        "lora_param": self.model.lora_param,
                    },
                    workdir / "lora.pt",
                )
            else:
                torch.save(
                    {"nn": self.model.state_dict(), "hparam": self.model.hparam},
                    workdir / "model_ft.pt",
                )



# -*- coding: utf-8 -*-
# Author: Nianze A. TAO (Omozawa SUENO)
"""
Define essential scorers.
"""
from typing import List, Callable, Union, Optional
import torch
import numpy as np
from torch import Tensor
from rdkit import RDLogger
from rdkit.Contrib.SA_Score import sascorer  # type: ignore
from rdkit.Chem import MolFromSmiles, QED
from scipy.stats import wasserstein_distance

RDLogger.DisableLog("rdApp.*")  # type: ignore


def smiles_valid(smiles: str) -> int:
    """
    Return the validity of a SMILES string.

    :param smiles: SMIlES string
    :type smiles: str
    :return: validity
    :rtype: int
    """
    return 1 if (MolFromSmiles(smiles) and smiles) else 0


def qed_score(smiles: str) -> float:
    """
    Return the quantitative estimate of drug-likeness score of a SMILES string.

    :param smiles: SMILES string
    :type smiles: str
    :return: QED score
    :rtype: float
    """
    return QED.qed(MolFromSmiles(smiles))


def sa_score(smiles: str) -> float:
    """
    Return the synthetic accessibility score of a SMILES string.

    :param smiles: SMILES string
    :type smiles: str
    :return: SA score
    :rtype: float
    """
    return sascorer.calculateScore(MolFromSmiles(smiles))


def spectra_wasserstein_score(
    spectrum_u: np.ndarray, spectrum_v: np.ndarray, x_axis: np.ndarray
) -> float:
    """
    Return the Wasserstein distance (earth mover's distance) between two
    continuous spectra scaled by the area under the first spectrum curve `spectrum_u`.

    :param spectrum_u: the reference spectrum
    :param spectrum_v: the
    :param x_axis: the shared x-axis of the spectra
    :type spectrum_u: numpy.ndarray
    :type spectrum_v: numpy.ndarray
    :type x_axis: numpy.ndarray
    :return: spectra Wasserstein score
    :rtype: float
    """
    assert spectrum_u.size == spectrum_v.size, "Spectra sizes should be matched."
    a = np.sqrt(np.trapezoid(spectrum_u, x_axis))
    return (wasserstein_distance(spectrum_u, spectrum_v) / a).item()


class Scorer:
    def __init__(
        self,
        scorers: List[Callable[[str], Union[int, float]]],
        score_criteria: List[Callable[[Union[int, float]], float]],
        vocab_keys: List[str],
        vocab_separator: str = "",
        valid_checker: Optional[Callable[[str], int]] = None,
        eta: float = 1e-2,
        name: str = "scorer",
    ) -> None:
        """
        Scorer class.
        e.g.

        ```python
        scorer = Scorer(
            scorers=[smiles_valid, qed_score],
            score_criteria=[lambda x: float(x == 1), lambda x: float(x > 0.5)],
            vocab_keys=VOCAB_KEYS,
        )
        ```

        :param scorers: a list of scorer(s)
        :param score_criteria: a list of score criterion (or criteria) in the same order of scorers
        :param vocab_keys: a list of (ordered) vocabulary
        :param vocab_separator: token separator; default is `""`
        :param valid_checker: a callable to check the validity of sequences; default is `None`
        :param eta: the coefficient to be multiplied to the loss
        :param name: the name of this scorer
        :type scorers: list
        :type score_criteria: list
        :type vocab_keys: list
        :type vocab_separator: str
        :type eta: float
        :type name: str
        :type valid_checker: typing.Callable | None
        """
        assert len(scorers) == len(
            score_criteria
        ), "The number of scores should match that of criteria."
        self.scorers = scorers
        self.score_criteria = score_criteria
        self.vocab_keys = vocab_keys
        self.vocab_separator = vocab_separator
        self.valid_checker = valid_checker
        self.eta = eta
        self.name = name

    def calc_score_loss(self, p: Tensor) -> Tensor:
        """
        Calculate the score loss.

        :param p: token probability distributions;  shape: (n_b, n_t, n_vocab)
        :type p: torch.Tensor
        :return: score loss;                        shape: ()
        :rtype: torch.Tensor
        """
        tokens = p.argmax(-1)
        e_k = torch.nn.functional.one_hot(tokens, len(self.vocab_keys)).float()
        seqs = [
            self.vocab_separator.join([self.vocab_keys[i] for i in j])
            .split("<start>" + self.vocab_separator)[-1]
            .split(self.vocab_separator + "<end>")[0]
            .replace("<pad>", "")
            for j in tokens
        ]
        valid = [
            1 if self.valid_checker is None else self.valid_checker(i) for i in seqs
        ]
        scores = [
            [
                1 if valid[j] == 0 else 1 - self.score_criteria[i](scorer(seq))
                for j, seq in enumerate(seqs)
            ]
            for i, scorer in enumerate(self.scorers)
        ]
        loss = (e_k * p).sum(2).mean(1) * p.new_tensor(scores).mean(0)
        return loss.mean()


if __name__ == "__main__":
    ...



# -*- coding: utf-8 -*-
# Author: Nianze A. TAO (Omozawa SUENO)
"""
Tokenise SMILES/SAFE/SELFIES/protein-sequence strings.
"""
import os
import re
from pathlib import Path
from typing import Any, List, Dict, Union, Callable
import torch
import torch.nn.functional as F
from torch import Tensor
from torch.utils.data import Dataset

__filedir__ = Path(__file__).parent

SMI_REGEX_PATTERN = (
    r"(\[|\]|H[e,f,g,s,o]?|"
    r"L[i,v,a,r,u]|"
    r"B[e,r,a,i,h,k]?|"
    r"C[l,a,r,o,u,d,s,n,e,m,f]?|"
    r"N[e,a,i,b,h,d,o,p]?|"
    r"O[s,g]?|S[i,c,e,r,n,m,b,g]?|"
    r"K[r]?|T[i,c,e,a,l,b,h,m,s]|"
    r"G[a,e,d]|R[b,u,h,e,n,a,f,g]|"
    r"Yb?|Z[n,r]|P[t,o,d,r,a,u,b,m]?|"
    r"F[e,r,l,m]?|M[g,n,o,t,c,d]|"
    r"A[l,r,s,g,u,t,c,m]|I[n,r]?|"
    r"W|X[e]|E[u,r,s]|U|D[b,s,y]|"
    r"b|c|n|o|s|p|"
    r"\(|\)|\.|=|#|-|\+|\\|\/|:|"
    r"~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
)
SEL_REGEX_PATTERN = r"(\[[^\]]+]|\.)"
AA_REGEX_PATTERN = r"(A|B|C|D|E|F|G|H|I|K|L|M|N|P|Q|R|S|T|V|W|Y|Z|-|.)"
smi_regex = re.compile(SMI_REGEX_PATTERN)
sel_regex = re.compile(SEL_REGEX_PATTERN)
aa_regex = re.compile(AA_REGEX_PATTERN)


def load_vocab(
    vocab_file: Union[str, Path],
) -> Dict[str, Union[int, List[str], Dict[str, int]]]:
    """
    Load vocabulary from source file.

    :param vocab_file: file that contains vocabulary
    :type vocab_file: str | pathlib.Path
    :return: {"vocab_keys": vocab_keys, "vocab_count": vocab_count, "vocab_dict": vocab_dict}
    :rtype: dict
    """
    with open(vocab_file, "r", encoding="utf-8") as f:
        lines = f.read().strip()
    vocab_keys = lines.split("\n")
    vocab_count = len(vocab_keys)
    vocab_dict = dict(zip(vocab_keys, range(vocab_count)))
    return {
        "vocab_keys": vocab_keys,
        "vocab_count": vocab_count,
        "vocab_dict": vocab_dict,
    }


_DEFUALT_VOCAB = load_vocab(__filedir__ / "vocab.txt")
VOCAB_KEYS: List[str] = _DEFUALT_VOCAB["vocab_keys"]
VOCAB_DICT: Dict[str, int] = _DEFUALT_VOCAB["vocab_dict"]
VOCAB_COUNT: int = _DEFUALT_VOCAB["vocab_count"]
AA_VOCAB_KEYS = (
    VOCAB_KEYS[0:3] + "A B C D E F G H I K L M N P Q R S T V W Y Z - .".split()
)
AA_VOCAB_COUNT = len(AA_VOCAB_KEYS)
AA_VOCAB_DICT = dict(zip(AA_VOCAB_KEYS, range(AA_VOCAB_COUNT)))


def smiles2vec(smiles: str) -> List[int]:
    """
    SMILES tokenisation using a dataset-independent regex pattern.

    :param smiles: SMILES string
    :type smiles: str
    :return: tokens w/o `<start>` and `<end>`
    :rtype: list
    """
    tokens = [token for token in smi_regex.findall(smiles)]
    return [VOCAB_DICT[token] for token in tokens]


def aa2vec(aa_seq: str) -> List[int]:
    """
    Protein sequence tokenisation using a dataset-independent regex pattern.

    :param aa_seq: protein (amino acid) sequence
    :type aa_seq: str
    :return: tokens w/o `<start>` and `<end>`
    :rtype: list
    """
    tokens = [token for token in aa_regex.findall(aa_seq)]
    return [AA_VOCAB_DICT[token] for token in tokens]


def split_selfies(selfies: str) -> List[str]:
    """
    SELFIES tokenisation.

    :param selfies: SELFIES string
    :type selfies: str
    :return: SELFIES vocab
    :rtype: list
    """
    return [token for token in sel_regex.findall(selfies)]


def smiles2token(smiles: str) -> Tensor:
    # start token: <start> = 1; end token: <esc> = 2
    return torch.tensor([1] + smiles2vec(smiles) + [2], dtype=torch.long)


def aa2token(aa_seq: str) -> Tensor:
    # start token: <start> = 1; end token: <end> = 2
    return torch.tensor([1] + aa2vec(aa_seq) + [2], dtype=torch.long)


def collate(batch: List[Dict[str, Tensor]]) -> Dict[str, Tensor]:
    """
    Padding the data in one batch into the same size.\n
    Should be passed to `~torch.utils.data.DataLoader` as `DataLoader(collate_fn=collate, ...)`.

    :param batch: a list of data (one batch)
    :type batch: list
    :return: batched {"token": token} or {"token": token, "value": value}
    :rtype: dict
    """
    token = [i["token"] for i in batch]
    if "MAX_PADDING_LENGTH" in os.environ:
        lmax = int(os.environ["MAX_PADDING_LENGTH"])
    else:
        lmax = max([len(w) for w in token])
    token = torch.cat(
        [F.pad(i, (0, lmax - len(i)), value=0)[None, :] for i in token], 0
    )
    out_dict = {"token": token}
    if "value" in batch[0]:
        out_dict["value"] = torch.cat([i["value"][None, :] for i in batch], 0)
    if "mask" in batch[0]:
        mask = [i["mask"] for i in batch]
        out_dict["mask"] = torch.cat(
            [F.pad(i, (0, lmax - len(i)), value=0)[None, :] for i in mask], 0
        )
    return out_dict


class CSVData(Dataset):
    def __init__(self, file: Union[str, Path]):
        """
        Define dataset stored in CSV file.

        :param file: dataset file name <file>
        :type file: str | pathlib.Path
        """
        super().__init__()
        with open(file, "r") as db:
            self.data = db.readlines()
        self.header_idx_dict: Dict[str, List[int]] = {}
        for key, i in enumerate(self.data[0].replace("\n", "").split(",")):
            if i in self.header_idx_dict:
                self.header_idx_dict[i].append(key)
            else:
                self.header_idx_dict[i] = [key]
        self.mapping = lambda x: x

    def __len__(self) -> int:
        return len(self.data) - 1

    def __getitem__(self, idx: Union[int, Tensor]) -> Dict[str, Tensor]:
        if torch.is_tensor(idx):
            idx = idx.tolist()
        # valid `idx` should start from 1 instead of 0
        data: List[str] = self.data[idx + 1].replace("\n", "").split(",")
        data_dict: Dict[str, List[str]] = {}
        for key in self.header_idx_dict:
            data_dict[key] = [data[i] for i in self.header_idx_dict[key]]
        return self.mapping(data_dict)

    def map(self, mapping: Callable[[Dict[str, List[str]]], Any]) -> None:
        """
        Pass a customised mapping function to transform the data entities to tensors.

        e.g.
        ```python
        import torch
        from bayesianflow_for_chem.data import smiles2token, CSVData


        def encode(x):
            return {
                "token": smiles2token(".".join(x["smiles"])),
                "value": torch.tensor([float(i) if i != "" else torch.inf for i in x["value"]]),
            }

        dataset = CSVData(...)
        dataset.map(encode)
        ```

        :param mapping: customised mapping function
        :type mapping: callable
        :return:
        :rtype: None
        """
        self.mapping = mapping


if __name__ == "__main__":
    ...



# -*- coding: utf-8 -*-
# Author: Nianze A. Tao (Omozawa Sueno)
"""
ChemBFN package.
"""
from . import data, tool, train, scorer
from .model import ChemBFN, MLP, EnsembleChemBFN

__all__ = ["data", "tool", "train", "scorer", "ChemBFN", "MLP", "EnsembleChemBFN"]
__version__ = "1.4.3"
__author__ = "Nianze A. Tao (Omozawa Sueno)"



# -*- coding: utf-8 -*-
# Author: Nianze A. Tao (Omozawa Sueno)
"""
Define Bayesian Flow Network for Chemistry (ChemBFN) model.
"""
from pathlib import Path
from copy import deepcopy
from typing import List, Tuple, Dict, Optional, Union, Callable
import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.functional import softmax, linear, dropout
from typing_extensions import Self


class Linear(nn.Linear):
    # Modified from https://github.com/microsoft/LoRA/blob/main/loralib/layers.py
    # We made it simpler and compatible with both `loralib` and `TorchScript`.
    def __init__(self, in_features: int, out_features: int, bias: bool = True, **kargs):
        """
        LoRA implemented in a dense layer.

        :param in_features: number of input features
        :param out_features: number of output features
        :param bias: whether to use additional bias
        :param device: device
        :param dtype: PyTorch data type
        :type in_features: int
        :type out_features: int
        :type bias: bool
        :type device: torch.device | str | None
        :type dtype: torch.dtype
        """
        nn.Linear.__init__(self, in_features, out_features, bias, **kargs)
        self.lora_enabled: bool = False
        self.lora_A: Optional[nn.Parameter] = None
        self.lora_B: Optional[nn.Parameter] = None
        self.scaling: Optional[float] = None
        self.lora_dropout: Optional[float] = None
        nn.Linear.reset_parameters(self)

    def enable_lora(
        self, r: int = 8, lora_alpha: int = 1, lora_dropout: float = 0.0
    ) -> None:
        """
        Enable LoRA parameters.

        :param r: rank
        :param lora_alpha: LoRA alpha value
        :param lora_dropout: dropout frequency in LoRA layer
        :type r: int
        :type lora_alpha: float
        :type lora_dropout: float
        :return:
        :rtype: None
        """
        assert r > 0, "Rank should be larger than 0."
        try:
            self.lora_A = nn.Parameter(self.weight.new_zeros((r, self.in_features)))
            self.lora_B = nn.Parameter(self.weight.new_zeros((self.out_features, r)))
        except NotImplementedError:
            self.lora_A = nn.Parameter(
                torch.zeros((r, self.in_features), device=self.weight.device)
            )
            self.lora_B = nn.Parameter(
                torch.zeros((self.out_features, r), device=self.weight.device)
            )
        self.scaling = lora_alpha / r
        self.lora_dropout = lora_dropout
        self.lora_enabled = True
        nn.init.kaiming_uniform_(self.lora_A, a=5**0.5)
        nn.init.zeros_(self.lora_B)
        self.weight.requires_grad_(False)

    def forward(self, x: Tensor) -> Tensor:
        result = linear(x, self.weight, self.bias)
        if self.lora_enabled and isinstance(self.lora_dropout, float):
            result += (
                dropout(x, self.lora_dropout, self.training)
                @ self.lora_A.transpose(0, 1)
                @ self.lora_B.transpose(0, 1)
            ) * self.scaling
        return result


def modulate(x: Tensor, shift: Tensor, scale: Tensor) -> Tensor:
    return x * (1 + scale) + shift


class RoPE(nn.Module):
    def __init__(self, channel: int = 512, num_head: int = 8) -> None:
        """
        Rotary position embedding block with XPOS method.

        :param channel: hidden layer features
        :param num_head: number of heads
        :type channel: int
        :type num_head: int
        """
        super().__init__()
        d = channel // num_head
        assert d % 2 == 0
        self.channel = channel
        i = torch.arange(0, d, 2)[None, :] / d
        theta_half = torch.pow(10000, -i)
        zeta_half = (i + 0.4) / 1.4
        theta, zeta = torch.zeros((1, d)), torch.zeros((1, d))
        theta[:, 0::2] = theta_half
        theta[:, 1::2] = theta_half
        zeta[:, 0::2] = zeta_half
        zeta[:, 1::2] = zeta_half
        self.register_buffer("theta", theta)
        self.register_buffer("zeta", zeta)

    def forward(self, size: int) -> Tuple[Tensor, Tensor, Tensor]:
        """
        :param size: maximum length of sequence in the batch
        :type size: int
        :return: cos part of position encoding;  shape: (1, 1, n_t, n_h) \n
                 sin part of position encoding;  shape: (1, 1, n_t, n_h) \n
                 scaling coefficients;           shape: (1, 1, n_t, n_h)
        :rtype: tuple
        """
        pos = torch.arange(size, device=self.theta.device)[:, None]
        cos, sin = torch.cos(pos * self.theta), torch.sin(pos * self.theta)
        zeta = torch.pow(self.zeta, pos / self.channel)
        return cos[None, None, ...], sin[None, None, ...], zeta[None, None, ...]


class Attention(nn.Module):
    def __init__(self, channel: int = 512, num_head: int = 8) -> None:
        """
        Multi-head self-attention block.

        :param channel: hidden layer features
        :param num_head: number of heads
        :type channel: int
        :type num_head: int
        """
        super().__init__()
        assert channel % num_head == 0
        self.d = channel // num_head  # head dimension
        self.nh = num_head  # number of heads
        self.tp = (2 * self.d) ** 0.5  # attention temperature
        self.qkv = Linear(channel, channel * 3)

    @staticmethod
    def _rotate(
        q: Tensor, k: Tensor, pe: Tuple[Tensor, Tensor, Tensor]
    ) -> Tuple[Tensor, Tensor]:
        q_rotate, k_rotate = torch.zeros_like(q), torch.zeros_like(k)
        q_rotate[..., 0::2] = -q[..., 1::2]
        q_rotate[..., 1::2] = q[..., 0::2]
        q = (q * pe[0] + q_rotate * pe[1]) * pe[2]
        k_rotate[..., 0::2] = -k[..., 1::2]
        k_rotate[..., 1::2] = k[..., 0::2]
        k = (k * pe[0] + k_rotate * pe[1]) / pe[2]
        return q, k

    def forward(
        self, x: Tensor, pe: Tuple[Tensor, Tensor, Tensor], mask: Optional[Tensor]
    ) -> Tensor:
        """
        :param x: output tensor;       shape: (n_b, n_t, n_f)
        :param pe: position encoding;  shape: (1, 1, n_t, n_h) * 3
        :param mask: attention mask;   shape: (1, n_b, n_t, n_t)
        :type x: torch.Tensor
        :type pe: tuple
        :type mask: torch.Tensor | None
        :return: attentioned output;   shape: (n_b, n_t, n_f)
        :rtype: torch.Tensor
        """
        n_b, n_t, _ = shape = x.shape
        split = (n_b, n_t, self.nh, self.d)
        q, k, v = self.qkv(x).chunk(3, -1)
        q = q.view(split).permute(2, 0, 1, 3).contiguous()
        k = k.view(split).permute(2, 0, 1, 3).contiguous()
        v = v.view(split).permute(2, 0, 1, 3).contiguous()
        q, k = self._rotate(q, k, pe)  # position embedding
        atten_out = nn.functional.scaled_dot_product_attention(
            q, k, v, mask, 0.0, False, scale=1 / self.tp
        )
        atten_out = atten_out.permute(1, 2, 0, 3).contiguous().view(shape)
        return atten_out

    def enable_lora(
        self, r: int = 4, lora_alpha: int = 1, lora_dropout: float = 0.0
    ) -> None:
        """
        Enable LoRA parameters.

        :param r: rank
        :param lora_alpha: LoRA alpha value
        :param lora_dropout: dropout frequency in LoRA layer
        :type r: int
        :type lora_alpha: float
        :type lora_dropout: float
        :return:
        :rtype: None
        """
        self.qkv.enable_lora(r, lora_alpha, lora_dropout)


class TransformerLayer(nn.Module):
    def __init__(
        self, channel: int = 512, num_head: int = 8, dropout: float = 0.01
    ) -> None:
        """
        Transfomer layer block.

        :param channel: hidden layer features
        :param num_head: number of attention heads
        :param dropout: dropout frequency
        :type channel: int
        :type num_head: int
        :type dropout: float
        """
        super().__init__()
        self.norm1 = nn.LayerNorm(channel, 1e-6, False)
        self.attention = Attention(channel, num_head)
        self.norm2 = nn.LayerNorm(channel, 1e-6, False)
        self.ffn = nn.Sequential(
            nn.Linear(channel, channel * 4),
            nn.SELU(),
            nn.Linear(channel * 4, channel),
            nn.Dropout(dropout),
        )
        self.adaln_modulation = nn.Sequential(nn.SELU(), Linear(channel, 6 * channel))
        # zero-out adaLN layer
        nn.init.constant_(self.adaln_modulation[1].weight, 0)
        nn.init.constant_(self.adaln_modulation[1].bias, 0)

    def forward(
        self,
        x: Tensor,
        pe: Tuple[Tensor, Tensor, Tensor],
        c: Tensor,
        mask: Optional[Tensor],
    ) -> Tensor:
        """
        :param x: input tensor;        shape: (n_b, n_t, n_f)
        :param pe: position encoding;  shape: (1, 1, n_t, n_h) * 3
        :param c: conditioning;        shape: (n_b, 1, n_f)
        :param mask: attention mask;   shape: (1, n_b, n_t, n_t)
        :type x: torch.Tensor
        :type pe: tuple
        :type c: torch.Tensor
        :type mask: torch.Tensor | None
        :return: output tensor;        shape: (n_b, n_t, n_f)
        :rtype: torch.Tensor
        """
        c = self.adaln_modulation(c)
        shift, scale, gate, shift_ffn, scale_ffn, gate_ffn = c.chunk(6, -1)
        x = x + gate * self.attention(modulate(self.norm1(x), shift, scale), pe, mask)
        x = x + gate_ffn * self.ffn(modulate(self.norm2(x), shift_ffn, scale_ffn))
        return x

    def enable_lora(
        self, r: int = 4, lora_alpha: int = 1, lora_dropout: float = 0.0
    ) -> None:
        """
        Enable LoRA parameters.

        :param r: rank
        :param lora_alpha: LoRA alpha value
        :param lora_dropout: dropout frequency in LoRA layer
        :type r: int
        :type lora_alpha: float
        :type lora_dropout: float
        :return:
        :rtype: None
        """
        self.attention.enable_lora(r, lora_alpha, lora_dropout)
        self.adaln_modulation[1].enable_lora(r, lora_alpha, lora_dropout)


class FinalLayer(nn.Module):
    def __init__(self, num_vocab: int, channel: int = 512) -> None:
        """
        The final layer of model.

        :param num_vocab: number of vocabulary
        :param channel: hidden layer features
        :type num_vocab: int
        :type channel: int
        """
        super().__init__()
        self.norm_final = nn.LayerNorm(channel, 1e-6, False)
        self.linear = Linear(channel, num_vocab)
        self.adaln_modulation = nn.Sequential(nn.SELU(), Linear(channel, 2 * channel))
        # zero-out this layer
        nn.init.constant_(self.linear.weight, 0)
        nn.init.constant_(self.linear.bias, 0)
        nn.init.constant_(self.adaln_modulation[-1].weight, 0)
        nn.init.constant_(self.adaln_modulation[-1].bias, 0)

    def forward(self, x: Tensor, c: Tensor, return_logits: bool = True) -> Tensor:
        """
        :param x: input tensor;                 shape: (n_b, n_t, n_f)
        :param c: conditioning;                 shape: (n_b, 1, n_f)
        :param return_logits: whether to return unnormalised output logits
        :type x: torch.Tensor
        :type c: torch.Tensor
        :type return_logits: bool
        :return: output logits (unnormalised);  shape: (n_b, n_t, n_vocab)
                 or token embeddings;           shape: (n_b, n_t, n_f)
        :rtype: torch.Tensor
        """
        shift, scale = self.adaln_modulation(c).chunk(2, -1)
        x = modulate(self.norm_final(x), shift, scale)
        if return_logits:
            return self.linear(x)
        return x

    def enable_lora(
        self, r: int = 4, lora_alpha: int = 1, lora_dropout: float = 0.0
    ) -> None:
        """
        Enable LoRA parameters.

        :param r: rank
        :param lora_alpha: LoRA alpha value
        :param lora_dropout: dropout frequency in LoRA layer
        :type r: int
        :type lora_alpha: float
        :type lora_dropout: float
        :return:
        :rtype: None
        """
        self.linear.enable_lora(r, lora_alpha, lora_dropout)
        self.adaln_modulation[1].enable_lora(r, lora_alpha, lora_dropout)


class ChemBFN(nn.Module):
    def __init__(
        self,
        num_vocab: int,
        channel: int = 512,
        num_layer: int = 12,
        num_head: int = 8,
        dropout: float = 0.01,
    ) -> None:
        r"""
        Bayesian Flow Network for Chemistry model representation.

        Enable semi-autoregressive sampling by setting
        `ChemBFN(...).semi_autoregressive = True`.

        :param num_vocab: number of vocabulary
        :param channel: hidden layer features
        :param num_layer: number of transformer layers
        :param num_head: number of heads
        :param dropout: dropout frequency
        :type num_vocab: int
        :type channel: int
        :type num_layer: int
        :type num_head: int
        :type dropout: float
        """
        super().__init__()
        self.K = num_vocab
        self.lora_enabled: bool = False
        self.semi_autoregressive: bool = False
        self.embedding = Linear(num_vocab, channel)
        self.time_embed = nn.Sequential(
            nn.Linear(1, channel // 2), nn.SELU(), nn.Linear(channel // 2, channel)
        )
        self.position = RoPE(channel, num_head)
        self.encoder_layers = nn.ModuleList(
            [TransformerLayer(channel, num_head, dropout) for _ in range(num_layer)]
        )
        self.final_layer = FinalLayer(num_vocab, channel)
        self.register_buffer("beta", torch.scalar_tensor(20.4054 / self.K))
        self.hparam = dict(
            num_vocab=num_vocab,
            channel=channel,
            num_layer=num_layer,
            num_head=num_head,
            dropout=dropout,
        )
        self.lora_param = {}

    def enable_lora(
        self, r: int = 4, lora_alpha: int = 1, lora_dropout: float = 0.0
    ) -> None:
        """
        Enable LoRA parameters.

        :param r: rank
        :param lora_alpha: LoRA alpha value
        :param lora_dropout: dropout frequency in LoRA layer
        :type r: int
        :type lora_alpha: float
        :type lora_dropout: float
        :return:
        :rtype: None
        """
        self.lora_enabled = True
        self.lora_param = dict(r=r, lora_alpha=lora_alpha, lora_dropout=lora_dropout)
        self.embedding.enable_lora(r, lora_alpha, lora_dropout)
        for layer in self.encoder_layers:
            layer.enable_lora(r, lora_alpha, lora_dropout)
        self.final_layer.enable_lora(r, lora_alpha, lora_dropout)

    def forward(
        self,
        x: Tensor,
        t: Tensor,
        mask: Optional[Tensor] = None,
        y: Optional[Tensor] = None,
    ) -> Tensor:
        """
        :param x: input probabilities;                       shape: (n_b, n_t, n_vocab)
        :param t: time;                                      shape: (n_b, 1, 1)
        :param mask: input mask;                             shape: (n_b, n_t, 1)
        :param y: conditioning vector;                       shape: (n_b, 1, n_f)
        :type x: torch.Tensor
        :type t: torch.Tensor
        :type mask: torch.Tensor | None
        :type y: torch.Tensor | None
        :return: probability distribution (before softmax);  shape: (n_b, n_t, n_vocab)
                 or token embeddings;                        shape: (n_b, n_t, n_f)
        :rtype: torch.Tensor
        """
        n_b, n_t, _ = x.shape
        c = self.time_embed(t)
        if y is not None:
            c += y
        pe = self.position(n_t)
        x = self.embedding(x)
        if self.semi_autoregressive:
            attn_mask = torch.tril(
                torch.ones((1, n_b, n_t, n_t), device=x.device), diagonal=0
            )
        elif mask is not None:
            attn_mask = mask.transpose(-2, -1).repeat(1, n_t, 1)[None, ...] != 0
        else:
            attn_mask = None
        for layer in self.encoder_layers:
            x = layer(x, pe, c, attn_mask)
        return self.final_layer(x, c, mask is None)

    def calc_beta(self, t: Tensor) -> Tensor:
        r"""
        Calculate beta(t) value.

        .. math::
        ```
        \begin{equation}
            \beta(t) = %
            -\frac{4\ln{(1 - t + te^{-\frac{K}{4}\beta(1)})}}{K}
        \end{equation}
        ```

        :param t: continuous time in [0, 1];  shape: (n_b, 1, 1)
        :type t: torch.Tensor
        :return: beta(t);                     shape: (n_b, 1, 1)
        :rtype: torch.Tensor
        """
        return -4 * (1 - t + t * (-self.K * self.beta / 4).exp()).log() / self.K

    def calc_discrete_alpha(self, t1: Tensor, t2: Tensor) -> Tensor:
        r"""
        Calculate alpha(i) value.

        .. math:: $\alpha(i) = \bate(t_{i}) - \beta(t_{i - 1})$

        :param t1: discrete time (i - 1) / n;  shape: (n_b, 1, 1)
        :param t2: discrete time i / n;        shape: (n_b, 1, 1)
        :type t1: torch.Tensor
        :type t2: torch.Tensor
        :return: alpha(i);                     shape: (n_b, 1, 1)
        :rtype: torch.Tensor
        """
        # assert t2 > t1
        return self.calc_beta(t2) - self.calc_beta(t1)

    def calc_cts_alpha(self, t: Tensor) -> Tensor:
        r"""
        Calculate alpha(t) / 2 value.

        .. math::
        ```
        \begin{equation}
            \alpha(t) = %
            \frac{d\beta(t)}{dt} = %
            \frac{4}{K}%
            \frac{1 - e^{-\frac{K}{4}\beta(1)}}%
            {1 - t + te^{-\frac{K}{4}\beta(1)}}
        \end{equation}
        ```

        :param t: continuous time in [0, 1];  shape: (n_b, 1, 1)
        :type t: torch.Tensor
        :return: alpha(t);                    shape: (n_b, 1, 1)
        :rtype: torch.Tensor
        """
        a = 1 - (-self.K * self.beta / 4).exp()
        b = 1 - t + t * (-self.K * self.beta / 4).exp()
        return 2 * a / b / self.K

    def discrete_output_distribution(
        self, theta: Tensor, t: Tensor, y: Optional[Tensor], w: Optional[float]
    ) -> Tensor:
        """
        :param theta: input distribution;     shape: (n_b, n_t, n_vocab)
        :param t: continuous time in [0, 1];  shape: (n_b, 1, 1)
        :param y: conditioning vector;        shape: (n_b, 1, n_f)
        :param w: guidance strength controlling the conditional generation
        :type theta: torch.Tensor
        :type t: torch.Tensor
        :type y: torch.Tensor | None
        :type w: float | None
        :return: output distribution;         shape: (n_b, n_t, n_vocab)
        :rtype: torch.Tensor
        """
        theta = 2 * theta - 1  # rescale to [-1, 1]
        if w is None:
            return softmax(self.forward(theta, t, None, y), -1)
        elif y is None:
            return softmax(self.forward(theta, t, None, None), -1)
        else:
            p_cond = self.forward(theta, t, None, y)
            p_uncond = self.forward(theta, t, None, None)
            return softmax((1 + w) * p_cond - w * p_uncond, -1)

    def cts_loss(
        self,
        x: Tensor,
        t: Tensor,
        y: Optional[Tensor],
        mask: Optional[Tensor] = None,
        return_output_dist: bool = False,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """
        Compute continuous-time loss.

        :param x: target data;                shape: (n_b, n_t)
        :param t: continuous time in [0, 1);  shape: (n_b, 1, 1)
        :param y: conditioning vector;        shape: (n_b, 1, n_f)
        :param mask: in-text mask;            shape: (n_b, n_t)
        :param return_output_dist: whether to return the output distribution
        :type x: torch.Tensor
        :type t: torch.Tensor
        :type y: torch.Tensor | None
        :type mask: torch.Tensor | None
        :type return_output_dist: bool
        :returns: continuous-time loss;       shape: () \n
                  output distribution;        shape: (n_b, n_t, n_vocab) or `None`
        :rtype: tuple
        """
        beta = self.calc_beta(t)  # shape: (n_b, 1, 1)
        e_x = nn.functional.one_hot(x, self.K).float()
        mu = beta * (self.K * e_x - 1)
        sigma = (beta * self.K).sqrt()
        theta = softmax(mu + sigma * torch.randn_like(mu), -1)
        if mask is not None:
            mask = mask[..., None]
            theta = e_x * mask + (1 - mask) * theta
        e_hat = self.discrete_output_distribution(theta, t, y, None)
        cts_loss = self.K * (e_x - e_hat).pow(2) * self.calc_cts_alpha(t)
        if return_output_dist:
            return cts_loss.mean(), e_hat
        return cts_loss.mean(), None

    @torch.inference_mode()
    def reconstruction_loss(self, x: Tensor, t: Tensor, y: Optional[Tensor]) -> Tensor:
        """
        Compute reconstruction loss.

        :param x: target data;                shape: (n_b, n_t)
        :param t: continuous time in [0, 1];  shape: (n_b, 1, 1)
        :param y: conditioning vector;        shape: (n_b, 1, n_f)
        :type x: torch.Tensor
        :type t: torch.Tensor
        :type y: torch.Tensor | None
        :return: reconstruction loss;         shape: ()
        :rtype: torch.Tensor
        """
        beta = self.calc_beta(t)
        mu = beta * (self.K * nn.functional.one_hot(x, self.K).float() - 1)
        sigma = (beta * self.K).sqrt()
        theta = softmax(mu + sigma * torch.randn_like(mu), -1)
        logits = self.forward(2 * theta - 1, t, None, y)
        # compute negative log probability
        x, logits = torch.broadcast_tensors(x[..., None], logits)
        return (-logits.gather(-1, x[..., :1]).squeeze(-1)).mean()

    @staticmethod
    def reshape_y(y: Tensor) -> Tensor:
        assert y.dim() <= 3  # this doesn't work if the model is frezen in JIT.
        if y.dim() == 2:
            return y[:, None, :]
        return y

    @torch.jit.export
    def sample(
        self,
        batch_size: int,
        sequence_size: int,
        y: Optional[Tensor],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Sample from a piror distribution.

        :param batch_size: batch size
        :param sequence_size: max sequence length
        :param y: conditioning vector;      shape: (n_b, 1, n_f) or (n_b, n_f)
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;      shape: (1, 1, n_vocab)
        :type batch_size: int
        :type sequence_size: int
        :type y: torch.Tensor | None
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :return: sampled token indices;     shape: (n_b, n_t) \n
                 entropy of the tokens;     shape: (n_b)
        :rtype: tuple
        """
        theta = (
            torch.ones((batch_size, sequence_size, self.K), device=self.beta.device)
            / self.K
        )
        if y is not None:
            y = self.reshape_y(y)
        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
            t = (i - 1).view(1, 1, 1).repeat(batch_size, 1, 1) / sample_step
            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
            if token_mask is not None:
                p = p.masked_fill_(token_mask, 0.0)
            alpha = self.calc_discrete_alpha(t, t + 1 / sample_step)
            e_k = nn.functional.one_hot(torch.argmax(p, -1), self.K).float()
            mu = alpha * (self.K * e_k - 1)
            sigma = (alpha * self.K).sqrt()
            theta = (mu + sigma * torch.randn_like(mu)).exp() * theta
            theta = theta / theta.sum(-1, True)
        t_final = torch.ones((batch_size, 1, 1), device=self.beta.device)
        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
        entropy = -(p * p.log()).sum(-1).mean(-1)
        if token_mask is not None:
            p = p.masked_fill_(token_mask, 0.0)
        return torch.argmax(p, -1), entropy

    @torch.jit.export
    def ode_sample(
        self,
        batch_size: int,
        sequence_size: int,
        y: Optional[Tensor],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
        temperature: float = 0.5,
    ) -> Tuple[Tensor, Tensor]:
        """
        ODE-based sampling.

        :param batch_size: batch size
        :param sequence_size: max sequence length
        :param y: conditioning vector;      shape: (n_b, 1, n_f) or (n_b, n_f)
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;      shape: (1, 1, n_vocab)
        :param temperature: sampling temperature
        :type batch_size: int
        :type sequence_size: int
        :type y: torch.Tensor | None
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :type temperature: float
        :return: sampled token indices;     shape: (n_b, n_t) \n
                 entropy of the tokens;     shape: (n_b)
        :rtype: tuple
        """
        z = torch.zeros((batch_size, sequence_size, self.K), device=self.beta.device)
        if y is not None:
            y = self.reshape_y(y)
        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
            t = (i - 1).view(1, 1, 1).repeat(batch_size, 1, 1) / sample_step
            theta = torch.softmax(z, -1)
            beta = self.calc_beta(t + 1 / sample_step)
            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
            if token_mask is not None:
                p = p.masked_fill_(token_mask, 0.0)
            u = torch.randn_like(z)
            z = (self.K * p - 1) * beta + (self.K * beta * temperature).sqrt() * u
        t_final = torch.ones((batch_size, 1, 1), device=self.beta.device)
        theta = torch.softmax(z, -1)
        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
        entropy = -(p * p.log()).sum(-1).mean(-1)
        if token_mask is not None:
            p = p.masked_fill_(token_mask, 0.0)
        return torch.argmax(p, -1), entropy

    @torch.jit.export
    def inpaint(
        self,
        x: Tensor,
        y: Optional[Tensor] = None,
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Molecule inpaint functionality.

        :param x: categorical indices of scaffold;  shape: (n_b, n_t)
        :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f)
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;              shape: (1, 1, n_vocab)
        :type x: torch.Tensor
        :type y: torch.Tensor | None
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :return: sampled token indices;             shape: (n_b, n_t) \n
                 entropy of the tokens;             shape: (n_b)
        :rtype: tuple
        """
        n_b, n_t = x.shape
        mask = (x != 0).float()[..., None]
        theta = torch.ones((n_b, n_t, self.K), device=x.device) / self.K
        x_onehot = nn.functional.one_hot(x, self.K) * mask
        theta = x_onehot + (1 - mask) * theta
        if y is not None:
            y = self.reshape_y(y)
        for i in torch.linspace(1, sample_step, sample_step, device=x.device):
            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
            if token_mask is not None:
                p = p.masked_fill_(token_mask, 0.0)
            alpha = self.calc_discrete_alpha(t, t + 1 / sample_step)
            e_k = nn.functional.one_hot(torch.argmax(p, -1), self.K).float()
            mu = alpha * (self.K * e_k - 1)
            sigma = (alpha * self.K).sqrt()
            theta = (mu + sigma * torch.randn_like(mu)).exp() * theta
            theta = theta / theta.sum(-1, True)
            theta = x_onehot + (1 - mask) * theta
        t_final = torch.ones((n_b, 1, 1), device=x.device)
        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
        entropy = -(p * p.log()).sum(-1).mean(-1)
        if token_mask is not None:
            p = p.masked_fill_(token_mask, 0.0)
        return torch.argmax(p, -1), entropy

    @torch.jit.export
    def ode_inpaint(
        self,
        x: Tensor,
        y: Optional[Tensor] = None,
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
        temperature: float = 0.5,
    ) -> Tuple[Tensor, Tensor]:
        """
        ODE inpainting.

        :param x: categorical indices of scaffold;  shape: (n_b, n_t)
        :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f)
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;              shape: (1, 1, n_vocab)
        :param temperature: sampling temperature
        :type x: torch.Tensor
        :type y: torch.Tensor | None
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :type temperature: float
        :return: sampled token indices;             shape: (n_b, n_t) \n
                 entropy of the tokens;             shape: (n_b)
        :rtype: tuple
        """
        n_b, n_t = x.shape
        mask = (x != 0).float()[..., None]
        x_onehot = nn.functional.one_hot(x, self.K) * mask
        z = torch.zeros((n_b, n_t, self.K), device=self.beta.device)
        if y is not None:
            y = self.reshape_y(y)
        for i in torch.linspace(1, sample_step, sample_step, device=self.beta.device):
            t = (i - 1).view(1, 1, 1).repeat(n_b, 1, 1) / sample_step
            theta = torch.softmax(z, -1)
            theta = x_onehot + (1 - mask) * theta
            beta = self.calc_beta(t + 1 / sample_step)
            p = self.discrete_output_distribution(theta, t, y, guidance_strength)
            if token_mask is not None:
                p = p.masked_fill_(token_mask, 0.0)
            u = torch.randn_like(z)
            z = (self.K * p - 1) * beta + (self.K * beta * temperature).sqrt() * u
        t_final = torch.ones((n_b, 1, 1), device=self.beta.device)
        theta = torch.softmax(z, -1)
        theta = x_onehot + (1 - mask) * theta
        p = self.discrete_output_distribution(theta, t_final, y, guidance_strength)
        entropy = -(p * p.log()).sum(-1).mean(-1)
        if token_mask is not None:
            p = p.masked_fill_(token_mask, 0.0)
        return torch.argmax(p, -1), entropy

    def inference(self, x: Tensor, mlp: nn.Module) -> Tensor:
        """
        Predict from SMILES tokens.

        :param x: input tokens;  shape: (n_b, n_t)
        :param mlp: MLP module
        :type x: torch.Tensor
        :type mlp: torch.nn.Module
        :return: output values;  shape: (n_b, n_task)
        :rtype: torch.Tensor
        """
        t = torch.ones((x.shape[0], 1, 1), device=x.device)
        mask = (x != 0).float()[..., None]
        theta = 2 * torch.nn.functional.one_hot(x, self.K).float() - 1
        z = self.forward(theta, t, mask, None)
        if self.semi_autoregressive:
            return mlp.forward(z[x == 2].view(z.shape[0], -1))
        return mlp.forward(z[::, 0])

    @classmethod
    def from_checkpoint(
        cls, ckpt: Union[str, Path], ckpt_lora: Union[str, Path, None] = None
    ) -> Self:
        """
        Load model weight from a checkpoint.

        :param ckpt: checkpoint file
        :param ckpt_lora: LoRA checkpoint file which is optional
        :type ckpt: str | pathlib.Path
        :type ckpt_lora: str | pathlib.Path | None
        :return: Bayesian Flow Network for Chemistry model
        :rtype: bayesianflow_for_chem.model.ChemBNF
        """
        with open(ckpt, "rb") as f:
            state = torch.load(f, "cpu", weights_only=True)
        nn, hparam = state["nn"], state["hparam"]
        model = cls(**hparam)
        model.load_state_dict(nn, False)
        if ckpt_lora:
            with open(ckpt_lora, "rb") as g:
                lora_state = torch.load(g, "cpu", weights_only=True)
            lora_nn, lora_param = lora_state["lora_nn"], lora_state["lora_param"]
            model.enable_lora(**lora_param)
            model.load_state_dict(lora_nn, False)
        return model


class MLP(nn.Module):
    def __init__(
        self, size: List[int], class_input: bool = False, dropout: float = 0.0
    ) -> None:
        """
        MLP module.
        e.g.

        ```python
        mlp = MLP(size=[512, 256, 1])
        mlp = MLP(size=[10, 256, 512], True)  # embedding 10 classes
        ```

        :param size: hidden feature sizes
        :param class_input: whether the input is class indices
        :param dropout: dropout frequency
        :type size: list
        :type class_input: bool
        :type dropout: float
        """
        super().__init__()
        assert len(size) >= 2
        self.class_input = class_input
        self.dropout = nn.Dropout(dropout if not class_input else 0.0)
        self.layers = nn.ModuleList(
            [nn.Linear(i, size[key + 1]) for key, i in enumerate(size[:-2])]
        )
        if class_input:
            self.layers[0] = nn.Embedding(size[0], size[1])
        self.layers.append(nn.Linear(size[-2], size[-1]))
        self.hparam = dict(size=size, class_input=class_input, dropout=dropout)

    def forward(self, x: Tensor) -> Tensor:
        """
        :param x: input tensor;  shape: (n_b, n_input)
        :return: output tensor;  shape: (n_b, n_output) if not class_input;
                                        (n_b, 1, n_output) if class_input
        :type x: torch.Tensor
        :rtype: torch.Tensor
        """
        x = self.dropout(x)
        if self.class_input:
            x = x.to(dtype=torch.long)
        for layer in self.layers[:-1]:
            x = torch.selu(layer.forward(x))
        return self.layers[-1](x)

    @classmethod
    def from_checkpoint(cls, ckpt: Union[str, Path], strict: bool = True) -> Self:
        """
        Load model weight from a checkpoint.

        :param ckpt: checkpoint file
        :param strict: whether to strictly match `state_dict`
        :type ckpt: str | pathlib.Path
        :type strict: bool
        :return: MLP
        :rtype: bayesianflow_for_chem.model.MLP
        """
        with open(ckpt, "rb") as f:
            state = torch.load(f, "cpu", weights_only=True)
        nn, hparam = state["nn"], state["hparam"]
        model = cls(**hparam)
        model.load_state_dict(nn, strict)
        return model


class EnsembleChemBFN(ChemBFN):
    """
    This module does not fully support `torch.jit.script`. We have `EnsembleChemBFN.jit()`
    method to JIT compile the submodels.
    `torch.compile()` is a better choice to compiling the whole model.
    """

    def __init__(
        self,
        base_model_path: Union[str, Path],
        lora_paths: Union[List[Union[str, Path]], Dict[str, Union[str, Path]]],
        cond_heads: Union[List[nn.Module], Dict[str, nn.Module]],
        adapter_weights: Optional[Union[List[float], Dict[str, float]]] = None,
        semi_autoregressive_flags: Optional[Union[List[bool], Dict[str, bool]]] = None,
    ) -> None:
        """
        Ensemble of ChemBFN models from LoRA checkpoints.

        :param base_model_path: base model checkpoint file
        :param lora_paths: a list of LoRA checkpoint files or a `dict` instance of these files
        :param cond_heads: a list of conditioning network heads or a `dict` instance of these networks
        :param adapter_weights: a list of weights of each LoRA finetuned model or a 'dict` instance of these weights; default is equally weighted
        :param semi_autoregressive_flags: a list of the semi-autoregressive behaviour states of each LoRA finetuned model or a `dict` instance of these states; default is all `False`
        :type base_model_path: str | pathlib.Path
        :type lora_paths: list | dict
        :type cond_heads: list | dict
        :type adapter_weights: list | dict | None
        :type semi_autoregressive_flags: list | dict | None
        """
        n = len(lora_paths)
        assert type(lora_paths) == type(
            cond_heads
        ), "`lora_paths` and `cond_heads` should have the same type!"
        assert n == len(
            cond_heads
        ), "`lora_paths` and `cond_heads` should have the same length!"
        if adapter_weights:
            assert type(lora_paths) == type(
                adapter_weights
            ), "`lora_paths` and `adapter_weights` should have the same type!"
            assert n == len(
                adapter_weights
            ), "`lora_paths` and `adapter_weights` should have the same length!"
        if semi_autoregressive_flags:
            assert type(lora_paths) == type(
                semi_autoregressive_flags
            ), "`lora_paths` and `semi_autoregressive_flags` should have the same type!"
            assert n == len(
                semi_autoregressive_flags
            ), "`lora_paths` and `semi_autoregressive_flags` should have the same length!"
        _label_is_dict = isinstance(lora_paths, dict)
        if isinstance(lora_paths, list):
            names = tuple(f"val_{i}" for i in range(n))
            lora_paths = dict(zip(names, lora_paths))
            cond_heads = dict(zip(names, cond_heads))
            if not adapter_weights:
                adapter_weights = (1 / n for _ in names)
            if not semi_autoregressive_flags:
                semi_autoregressive_flags = (False for _ in names)
            adapter_weights = dict(zip(names, adapter_weights))
            semi_autoregressive_flags = dict(zip(names, semi_autoregressive_flags))
        else:
            names = tuple(lora_paths.keys())
            if not adapter_weights:
                adapter_weights = dict(zip(names, (1 / n for _ in names)))
            if not semi_autoregressive_flags:
                semi_autoregressive_flags = dict(zip(names, (False for _ in names)))
        base_model = ChemBFN.from_checkpoint(base_model_path)
        models = dict(zip(names, (deepcopy(base_model.eval()) for _ in names)))
        for k in names:
            with open(lora_paths[k], "rb") as f:
                state = torch.load(f, "cpu", weights_only=True)
            lora_nn, lora_param = state["lora_nn"], state["lora_param"]
            models[k].enable_lora(**lora_param)
            models[k].load_state_dict(lora_nn, False)
            models[k].semi_autoregressive = semi_autoregressive_flags[k]
        super().__init__(**base_model.hparam)
        self.cond_heads = nn.ModuleDict(cond_heads)
        self.models = nn.ModuleDict(models)
        self.adapter_weights = adapter_weights
        self._label_is_dict = _label_is_dict  # flag
        # ------- remove unnecessary submodules -------
        self.embedding = None
        self.time_embed = None
        self.position = None
        self.encoder_layers = None
        self.final_layer = None
        self.__delattr__("embedding")
        self.__delattr__("time_embed")
        self.__delattr__("position")
        self.__delattr__("encoder_layers")
        self.__delattr__("final_layer")
        # ------- remove unused attributes -------
        self.__delattr__("semi_autoregressive")
        self.__delattr__("lora_enabled")
        self.__delattr__("lora_param")
        self.__delattr__("hparam")

    def construct_y(
        self, c: Union[List[Tensor], Dict[str, Tensor]]
    ) -> Dict[str, Tensor]:
        assert (
            isinstance(c, dict) is self._label_is_dict
        ), f"`c` should be a {'`dict` instance' if self._label_is_dict else '`list` instance'} but got {type(c)} instand."
        out: Dict[str, Tensor] = {}
        if isinstance(c, list):
            c = dict(zip([f"val_{i}" for i in range(len(c))], c))
        for name, model in self.cond_heads.items():
            y = model.forward(c[name])
            if y.dim() == 2:
                y = y[:, None, :]
            out[name] = y
        return out

    def discrete_output_distribution(
        self, theta: Tensor, t: Tensor, y: Dict[str, Tensor], w: float
    ) -> Tensor:
        """
        :param theta: input distribution;          shape: (n_b, n_t, n_vocab)
        :param t: continuous time in [0, 1];       shape: (n_b, 1, 1)
        :param y: a dict of conditioning vectors;  shape: (n_b, 1, n_f) * n_h
        :param w: guidance strength controlling the conditional generation
        :type theta: torch.Tensor
        :type t: torch.Tensor
        :type y: dict
        :type w: float
        :return: output distribution;              shape: (n_b, n_t, n_vocab)
        :rtype: torch.Tensor
        """
        theta = 2 * theta - 1  # rescale to [-1, 1]
        p_uncond, p_cond = torch.zeros_like(theta), torch.zeros_like(theta)
        # Q: Why not use `torch.vmap`? It's faster than doing the loop, isn't it?
        #
        # A: We have quite a few reasons to avoid using `vmap`:
        #    1. JIT doesn't support vmap;
        #    2. It's harder to switch on/off semi-autroregssive behaviours for individual
        #       models when all models are stacked into one (we have a solution but it's not
        #       that elegant);
        #    3. We just found that the result from vmap was not identical to doing the loop;
        #    4. vmap requires all models have the same size but it's not always that case
        #       since we sometimes use different ranks of LoRA in finetuning.
        for name, model in self.models.items():
            p_uncond_ = model.forward(theta, t, None, None)
            p_uncond += p_uncond_ * self.adapter_weights[name]
            p_cond_ = model.forward(theta, t, None, y[name])
            p_cond += p_cond_ * self.adapter_weights[name]
        return softmax((1 + w) * p_cond - w * p_uncond, -1)

    @staticmethod
    def reshape_y(y: Dict[str, Tensor]) -> Dict[str, Tensor]:
        for k in y:
            assert y[k].dim() <= 3
            if y[k].dim() == 2:
                y[k] = y[k][:, None, :]
        return y

    @torch.inference_mode()
    def sample(
        self,
        batch_size: int,
        sequence_size: int,
        conditions: Union[List[Tensor], Dict[str, Tensor]],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Sample from a piror distribution.

        :param batch_size: batch size
        :param sequence_size: max sequence length
        :param conditions: guidance conditions;  shape: (n_b, n_c) * n_h
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;           shape: (1, 1, n_vocab)
        :type batch_size: int
        :type sequence_size: int
        :type conditions: list | dict
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :return: sampled token indices;          shape: (n_b, n_t) \n
                 entropy of the tokens;          shape: (n_b)
        :rtype: tuple
        """
        y = self.construct_y(conditions)
        return super().sample(
            batch_size, sequence_size, y, sample_step, guidance_strength, token_mask
        )

    @torch.inference_mode()
    def ode_sample(
        self,
        batch_size: int,
        sequence_size: int,
        conditions: Union[List[Tensor], Dict[str, Tensor]],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
        temperature: float = 0.5,
    ) -> Tuple[Tensor, Tensor]:
        """
        ODE-based sampling.

        :param batch_size: batch size
        :param sequence_size: max sequence length
        :param conditions: conditioning vector;  shape: (n_b, n_c) * n_h
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;           shape: (1, 1, n_vocab)
        :param temperature: sampling temperature
        :type batch_size: int
        :type sequence_size: int
        :type conditions: list | dict
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :type temperature: float
        :return: sampled token indices;          shape: (n_b, n_t) \n
                 entropy of the tokens;          shape: (n_b)
        :rtype: tuple
        """
        y = self.construct_y(conditions)
        return super().ode_sample(
            batch_size,
            sequence_size,
            y,
            sample_step,
            guidance_strength,
            token_mask,
            temperature,
        )

    @torch.inference_mode()
    def inpaint(
        self,
        x: Tensor,
        conditions: Union[List[Tensor], Dict[str, Tensor]],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Molecule inpaint functionality.

        :param x: categorical indices of scaffold;  shape: (n_b, n_t)
        :param conditions: conditioning vector;     shape: (n_b, n_c) * n_h
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;              shape: (1, 1, n_vocab)
        :type x: torch.Tensor
        :type conditions: list | dict
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :return: sampled token indices;             shape: (n_b, n_t) \n
                 entropy of the tokens;             shape: (n_b)
        :rtype: tuple
        """
        y = self.construct_y(conditions)
        return super().inpaint(x, y, sample_step, guidance_strength, token_mask)

    @torch.inference_mode()
    def ode_inpaint(
        self,
        x: Tensor,
        conditions: Union[List[Tensor], Dict[str, Tensor]],
        sample_step: int = 100,
        guidance_strength: float = 4.0,
        token_mask: Optional[Tensor] = None,
        temperature: float = 0.5,
    ) -> Tuple[Tensor, Tensor]:
        """
        ODE inpainting.

        :param x: categorical indices of scaffold;  shape: (n_b, n_t)
        :param conditions: conditioning vector;     shape: (n_b, n_c) * n_h
        :param sample_step: number of sampling steps
        :param guidance_strength: strength of conditional generation. It is not used if y is null.
        :param token_mask: token mask;              shape: (1, 1, n_vocab)
        :param temperature: sampling temperature
        :type x: torch.Tensor
        :type conditions: list | dict
        :type sample_step: int
        :type guidance_strength: float
        :type token_mask: torch.Tensor | None
        :type temperature: float
        :return: sampled token indices;             shape: (n_b, n_t) \n
                 entropy of the tokens;             shape: (n_b)
        :rtype: tuple
        """
        y = self.construct_y(conditions)
        return super().ode_inpaint(
            x, y, sample_step, guidance_strength, token_mask, temperature
        )

    def quantise(
        self, quantise_method: Optional[Callable[[ChemBFN], None]] = None
    ) -> None:
        """
        Quantise the submodels. \n
        This method should be called, if necessary, before `torch.compile()`.

        :param quantise_method: quantisation method; default is `bayesianflow_for_chem.tool.quantise_model_`
        :type quantise_method: callable | None
        :return:
        :rtype: None
        """
        if quantise_method is None:
            from bayesianflow_for_chem.tool import quantise_model_

            quantise_method = quantise_model_
        for _, v in self.models.items():
            quantise_method(v)

    def jit(self, freeze: bool = False) -> None:
        """
        JIT compile the submodels. \n
        This method should be called, if necessary, before `quantise()` method is called if applied.

        :param freeze: whether to freeze the submodels; default is `False`. If set to `True` this
                       method should be called before moving the model to a different device.
        :type freeze: bool
        :return:
        :rtype: None
        """
        for k, v in self.models.items():
            self.models[k] = torch.jit.script(v)
            if freeze:
                self.models[k] = torch.jit.freeze(
                    self.models[k], ["semi_autoregressive"]
                )

    @torch.jit.ignore
    def forward(self, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")

    def cts_loss(self, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")

    def reconstruction_loss(self, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")

    def enable_lora(self, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")

    def inference(self, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")

    @classmethod
    def from_checkpoint(cls, *_, **__) -> None:
        """
        Don't use this method!
        """
        raise NotImplementedError("There's nothing here!")


if __name__ == "__main__":
    ...



# -*- coding: utf-8 -*-
# Author: Nianze A. TAO (Omozawa SUENO)
import os
import re
from pathlib import Path
from shutil import rmtree
from setuptools import setup, find_packages

init_file = Path("bayesianflow_for_chem") / "__init__.py"

with open(init_file, mode="r", encoding="utf-8") as f:
    lines = f.readlines()
    for line in lines:
        if "__version__" in line:
            version = re.findall(r"[0-9]+\.[0-9]+\.[0-9]+", line)
            if len(version) != 0:
                version = version[0]
                print("version:", version)
                break

with open("README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="bayesianflow_for_chem",
    version=version,
    url="https://augus1999.github.io/bayesian-flow-network-for-chemistry/",
    description="Bayesian flow network framework for Chemistry",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="AGPL-3.0-or-later",
    license_files=["LICEN[CS]E*"],
    package_dir={"bayesianflow_for_chem": "bayesianflow_for_chem"},
    package_data={"bayesianflow_for_chem": ["./*.txt", "./*.py"]},
    include_package_data=True,
    author="Nianze A. Tao",
    author_email="tao-nianze@hiroshima-u.ac.jp",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "rdkit>=2025.3.5",
        "torch>=2.8.0",
        "torchao>=0.12",
        "numpy>=2.3.2",
        "scipy>=1.16.1",
        "loralib>=0.1.2",
        "lightning>=2.5.3",
        "scikit-learn>=1.7.1",
        "typing_extensions>=4.14.1",
    ],
    project_urls={
        "Source": "https://github.com/Augus1999/bayesian-flow-network-for-chemistry"
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords=["Chemistry", "CLM", "ChemBFN"],
)

if os.path.exists("build"):
    rmtree("build")
if os.path.exists("bayesianflow_for_chem.egg-info"):
    rmtree("bayesianflow_for_chem.egg-info")



# -*- coding: utf-8 -*-
# Author: Nianze A. TAO (Omozawa SUENO)
"""
Essential tools.
"""
import csv
import random
import warnings
from pathlib import Path
from typing import List, Dict, Tuple, Union, Optional
import torch
import numpy as np
from torch import cuda, Tensor, softmax
from torch.utils.data import DataLoader
from rdkit.Chem import (
    rdDetermineBonds,
    GetFormalCharge,
    MolFromXYZBlock,
    MolFromSmiles,
    MolToSmiles,
    CanonSmiles,
    AllChem,
    AddHs,
    Mol,
)
from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles  # type: ignore
from sklearn.metrics import (
    roc_auc_score,
    auc,
    precision_recall_curve,
    r2_score,
    mean_absolute_error,
    root_mean_squared_error,
)
from .data import VOCAB_KEYS
from .model import ChemBFN, MLP, EnsembleChemBFN


def _find_device() -> torch.device:
    if cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")


@torch.no_grad()
def test(
    model: ChemBFN,
    mlp: MLP,
    data: DataLoader,
    mode: str = "regression",
    device: Union[str, torch.device, None] = None,
) -> Dict[str, float]:
    """
    Test the trained network.

    :param model: pretrained ChemBFN model
    :param mlp: trained MLP model for testing
    :param data: DataLoader instance
    :param mode: testing mode chosen from `'regression'` and `'classification'`
    :param device: hardware accelerator
    :type model: bayesianflow_for_chem.model.ChemBFN
    :type mlp: bayesianflow_for_chem.model.MLP
    :type data: torch.utils.data.DataLoader
    :type mode: str
    :type device: str | torch.device | None
    :return: MAE & RMSE & R^2 / ROC-AUC & PRC-AUC
    :rtype: dict
    """
    if device is None:
        device = _find_device()
    model.to(device).eval()
    mlp.to(device).eval()
    predict_y, label_y = [], []
    for d in data:
        x, y = d["token"].to(device), d["value"]
        label_y.append(y)
        if mode == "regression":
            y_hat = model.inference(x, mlp)
        if mode == "classification":
            n_b, n_y = y.shape
            y_hat = softmax(model.inference(x, mlp).reshape(n_b * n_y, -1), -1)
            y_hat = y_hat.reshape(n_b, -1)
        predict_y.append(y_hat.detach().to("cpu"))
    predict_y, label_y = torch.cat(predict_y, 0), torch.cat(label_y, 0).split(1, -1)
    if mode == "regression":
        predict_y = [
            predict[label_y[i] != torch.inf]
            for (i, predict) in enumerate(predict_y.split(1, -1))
        ]
        label_y = [label[label != torch.inf] for label in label_y]
        y_zipped = list(zip(label_y, predict_y))
        mae = [mean_absolute_error(label, predict) for (label, predict) in y_zipped]
        rmse = [
            root_mean_squared_error(label, predict) for (label, predict) in y_zipped
        ]
        r2 = [r2_score(label, predict) for (label, predict) in y_zipped]
        return {"MAE": mae, "RMSE": rmse, "R^2": r2}
    if mode == "classification":
        n_c = len(label_y)
        predict_y = predict_y.chunk(n_c, -1)
        y_zipped = list(zip(label_y, predict_y))
        roc_auc = [
            roc_auc_score(
                label.flatten(),
                predict[:, 1] if predict.shape[-1] == 2 else predict,
                multi_class="raise" if predict.shape[-1] == 2 else "ovo",
                labels=None if predict.shape[-1] == 2 else range(predict.shape[-1]),
            )
            for (label, predict) in y_zipped
        ]
        try:
            prc = [
                precision_recall_curve(label.flatten(), predict[:, 1])[:2]
                for (label, predict) in y_zipped
            ]
            prc_auc = [auc(recall, precision) for (precision, recall) in prc]
        except ValueError:
            prc_auc = []
        return {"ROC-AUC": roc_auc, "PRC-AUC": prc_auc}


def split_dataset(
    file: Union[str, Path], split_ratio: List[int] = [8, 1, 1], method: str = "random"
) -> None:
    """
    Split a dataset.

    :param file: dataset file <file>
    :param split_ratio: traing-testing-validation ratio
    :param method: chosen from `'random'` and `'scaffold'`
    :type file: str | pathlib.Path
    :type split_ratio: list
    :type method: str
    :return:
    :rtype: None
    """
    if isinstance(file, Path):
        file = file.__str__()
    assert file.endswith(".csv")
    assert len(split_ratio) == 3
    assert method in ("random", "scaffold")
    with open(file, "r") as f:
        data = list(csv.reader(f))
    header = data[0]
    raw_data = data[1:]
    smiles_idx = []  # only first index will be used
    for key, h in enumerate(header):
        if "smiles" in h.lower():
            smiles_idx.append(key)
    assert len(smiles_idx) > 0
    data_len = len(raw_data)
    train_ratio = split_ratio[0] / sum(split_ratio)
    test_ratio = sum(split_ratio[:2]) / sum(split_ratio)
    train_idx, test_idx = int(data_len * train_ratio), int(data_len * test_ratio)
    if method == "random":
        random.shuffle(raw_data)
        train_set = raw_data[:train_idx]
        test_set = raw_data[train_idx:test_idx]
        val_set = raw_data[test_idx:]
    if method == "scaffold":
        scaffolds: Dict[str, List] = {}
        for key, d in enumerate(raw_data):
            # compute Bemis-Murcko scaffold
            if len(smiles_idx) > 1:
                warnings.warn(
                    "\033[32;1m"
                    f"We found {len(smiles_idx)} SMILES strings in a row!"
                    " Only the first SMILES will be used to compute the molecular scaffold."
                    "\033[0m",
                    stacklevel=2,
                )
            try:
                scaffold = MurckoScaffoldSmiles(d[smiles_idx[0]])
                if scaffold in scaffolds:
                    scaffolds[scaffold].append(key)
                else:
                    scaffolds[scaffold] = [key]
            except ValueError:  # do nothing when SMILES is not valid
                ...
        scaffolds = {key: sorted(value) for key, value in scaffolds.items()}
        train_set, test_set, val_set = [], [], []
        for idxs in scaffolds.values():
            if len(train_set) + len(idxs) > train_idx:
                if len(train_set) + len(test_set) + len(idxs) > test_idx:
                    val_set += [raw_data[i] for i in idxs]
                else:
                    test_set += [raw_data[i] for i in idxs]
            else:
                train_set += [raw_data[i] for i in idxs]
    with open(file.replace(".csv", "_train.csv"), "w", newline="") as ftr:
        writer = csv.writer(ftr)
        writer.writerows([header] + train_set)
    with open(file.replace(".csv", "_test.csv"), "w", newline="") as fte:
        writer = csv.writer(fte)
        writer.writerows([header] + test_set)
    with open(file.replace(".csv", "_val.csv"), "w", newline="") as fva:
        writer = csv.writer(fva)
        writer.writerows([header] + val_set)


@torch.no_grad()
def sample(
    model: Union[ChemBFN, EnsembleChemBFN],
    batch_size: int,
    sequence_size: int,
    sample_step: int = 100,
    y: Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]] = None,
    guidance_strength: float = 4.0,
    device: Union[str, torch.device, None] = None,
    vocab_keys: List[str] = VOCAB_KEYS,
    seperator: str = "",
    method: str = "BFN",
    allowed_tokens: Union[str, List[str]] = "all",
    sort: bool = False,
) -> List[str]:
    """
    Sampling.

    :param model: trained ChemBFN model
    :param batch_size: batch size
    :param sequence_size: max sequence length
    :param sample_step: number of sampling steps
    :param y: conditioning vector;             shape: (n_b, 1, n_f) or (n_b, n_f) \n
              or a list/`dict` of conditions;  shape: (n_b, n_c) * n_h

    :param guidance_strength: strength of conditional generation. It is not used if y is null.
    :param device: hardware accelerator
    :param vocab_keys: a list of (ordered) vocabulary
    :param separator: token separator; default is `""`
    :param method: sampling method chosen from `"ODE:x"` or `"BFN"` where `x` is the value of sampling temperature; default is `"BFN"`
    :param allowed_tokens: a list of allowed tokens; default is `"all"`
    :param sort: whether to sort the samples according to entropy values; default is `False`
    :type model: bayesianflow_for_chem.model.ChemBFN | bayesianflow_for_chem.model.EnsembleChemBFN
    :type batch_size: int
    :type sequence_size: int
    :type sample_step: int
    :type y: torch.Tensor | list | dict | None
    :type guidance_strength: float
    :type device: str | torch.device | None
    :type vocab_keys: list
    :type separator: str
    :type method: str
    :type allowed_tokens: str | list
    :type sort: bool
    :return: a list of generated molecular strings
    :rtype: list
    """
    assert method.split(":")[0].lower() in ("ode", "bfn")
    if isinstance(model, EnsembleChemBFN):
        assert y is not None, "conditioning is required while using an ensemble model."
        assert isinstance(y, list) or isinstance(y, dict)
    else:
        assert isinstance(y, Tensor) or y is None
    if device is None:
        device = _find_device()
    model.to(device).eval()
    if y is not None:
        if isinstance(y, Tensor):
            y = y.to(device)
        elif isinstance(y, list):
            y = [i.to(device) for i in y]
        elif isinstance(y, dict):
            y = {k: v.to(device) for k, v in y.items()}
        else:
            raise NotImplementedError
    if isinstance(allowed_tokens, list):
        token_mask = [0 if i in allowed_tokens else 1 for i in vocab_keys]
        token_mask = torch.tensor([[token_mask]], dtype=torch.bool).to(device)
    else:
        token_mask = None
    if "ode" in method.lower():
        tp = float(method.split(":")[-1])
        assert tp > 0, "Sampling temperature should be higher than 0."
        tokens, entropy = model.ode_sample(
            batch_size, sequence_size, y, sample_step, guidance_strength, token_mask, tp
        )
    else:
        tokens, entropy = model.sample(
            batch_size, sequence_size, y, sample_step, guidance_strength, token_mask
        )
    if sort:
        sorted_idx = entropy.argsort(stable=True)
        tokens = tokens[sorted_idx]
    return [
        seperator.join([vocab_keys[i] for i in j])
        .split("<start>" + seperator)[-1]
        .split(seperator + "<end>")[0]
        .replace("<pad>", "")
        for j in tokens
    ]


@torch.no_grad()
def inpaint(
    model: Union[ChemBFN, EnsembleChemBFN],
    x: Tensor,
    sample_step: int = 100,
    y: Optional[Union[Tensor, Dict[str, Tensor], List[Tensor]]] = None,
    guidance_strength: float = 4.0,
    device: Union[str, torch.device, None] = None,
    vocab_keys: List[str] = VOCAB_KEYS,
    separator: str = "",
    method: str = "BFN",
    allowed_tokens: Union[str, List[str]] = "all",
    sort: bool = False,
) -> List[str]:
    """
    Inpaint (context guided) sampling.

    :param model: trained ChemBFN model
    :param x: categorical indices of scaffold;  shape: (n_b, n_t)
    :param sample_step: number of sampling steps
    :param y: conditioning vector;              shape: (n_b, 1, n_f) or (n_b, n_f) \n
              or a list/`dict` of conditions;   shape: (n_b, n_c) * n_h

    :param guidance_strength: strength of conditional generation. It is not used if y is null.
    :param device: hardware accelerator
    :param vocab_keys: a list of (ordered) vocabulary
    :param separator: token separator; default is `""`
    :param method: sampling method chosen from `"ODE:x"` or `"BFN"` where `x` is the value of sampling temperature; default is `"BFN"`
    :param allowed_tokens: a list of allowed tokens; default is `"all"`
    :param sort: whether to sort the samples according to entropy values; default is `False`
    :type model: bayesianflow_for_chem.model.ChemBFN | bayesianflow_for_chem.model.EnsembleChemBFN
    :type x: torch.Tensor
    :type sample_step: int
    :type y: torch.Tensor | list | dict | None
    :type guidance_strength: float
    :type device: str | torch.device | None
    :type vocab_keys: list
    :type separator: str
    :type method: str
    :type allowed_tokens: str | list
    :type sort: bool
    :return: a list of generated molecular strings
    :rtype: list
    """
    assert method.split(":")[0].lower() in ("ode", "bfn")
    if isinstance(model, EnsembleChemBFN):
        assert y is not None, "conditioning is required while using an ensemble model."
        assert isinstance(y, list) or isinstance(y, dict)
    else:
        assert isinstance(y, Tensor) or y is None
    if device is None:
        device = _find_device()
    model.to(device).eval()
    x = x.to(device)
    if y is not None:
        if isinstance(y, Tensor):
            y = y.to(device)
        elif isinstance(y, list):
            y = [i.to(device) for i in y]
        elif isinstance(y, dict):
            y = {k: v.to(device) for k, v in y.items()}
        else:
            raise NotImplementedError
    if isinstance(allowed_tokens, list):
        token_mask = [0 if i in allowed_tokens else 1 for i in vocab_keys]
        token_mask = torch.tensor([[token_mask]], dtype=torch.bool).to(device)
    else:
        token_mask = None
    if "ode" in method.lower():
        tp = float(method.split(":")[-1])
        assert tp > 0, "Sampling temperature should be higher than 0."
        tokens, entropy = model.ode_inpaint(
            x, y, sample_step, guidance_strength, token_mask, tp
        )
    else:
        tokens, entropy = model.inpaint(
            x, y, sample_step, guidance_strength, token_mask
        )
    if sort:
        sorted_idx = entropy.argsort(stable=True)
        tokens = tokens[sorted_idx]
    return [
        separator.join([vocab_keys[i] for i in j])
        .split("<start>" + separator)[-1]
        .split(separator + "<end>")[0]
        .replace("<pad>", "")
        for j in tokens
    ]


def quantise_model_(model: ChemBFN) -> None:
    """
    In-place dynamic quantisation of the trained model to `int8` data type. \n
    Due to some limitations of `torchao` module, not all layers will be quantised.

    :param model: trained ChemBFN model
    :type model: bayesianflow_for_chem.model.ChemBFN
    :return:
    :rtype: None
    """
    from torchao.quantization.quant_api import (
        quantize_,
        Int8DynamicActivationInt8WeightConfig,
    )

    quantize_(model, Int8DynamicActivationInt8WeightConfig())


def build_uv_vis_sepctrum(
    etoscs: np.ndarray, etenergies: np.ndarray, lambdas: np.ndarray
) -> np.ndarray:
    """
    Build UV/Vis spectrum from calculated electron transtion energies and oscillator strengths. \n
    This function follows the GaussView style: https://gaussian.com/uvvisplot/.

    :param etoscs: oscillator strengths
    :param etenergies: transtion energies
    :param lambdas: wavelengths
    :type etoscs: numpy.ndarray
    :type etenergies: numpy.ndarray
    :type lambdas: numpy.ndarray
    :return: absorption coefficient corrospending to the wavelengths
    :rtype: numpy.ndarray
    """
    return (
        etoscs[:, None]
        * np.exp(
            -np.pow((1 / lambdas[None, :] - etenergies[:, None] / 45.5634) * 3099.6, 2)
        )
    ).sum(0) * 40489.99421


class GeometryConverter:
    """
    Converting between different 2D/3D molecular representations.
    """

    @staticmethod
    def _xyz2mol(symbols: List[str], coordinates: np.ndarray) -> Mol:
        xyz_block = [str(len(symbols)), ""]
        r = coordinates
        for i, atom in enumerate(symbols):
            xyz_block.append(f"{atom} {r[i][0]:.10f} {r[i][1]:.10f} {r[i][2]:.10f}")
        return MolFromXYZBlock("\n".join(xyz_block))

    @staticmethod
    def smiles2cartesian(
        smiles: str,
        num_conformers: int = 250,
        rdkit_ff_type: str = "MMFF",
        refine_with_crest: bool = False,
        spin: float = 0.0,
    ) -> Tuple[List[str], np.ndarray]:
        """
        Guess the 3D geometry from SMILES string via MMFF conformer search.

        :param smiles: a valid SMILES string
        :param num_conformers: number of initial conformers
        :param rdkit_ff_type: force field type chosen in `'MMFF'` and `'UFF'`
        :param refine_with_crest: find the best conformer via CREST
        :param spin: total spin; only required when `refine_with_crest=True`
        :type smiles: str
        :type num_conformers: int
        :type rdkit_ff_type: str
        :type refine_with_crest: bool
        :type spin: float
        :return: atomic symbols \n
                 cartesian coordinates;  shape: (n_a, 3)
        :rtype: tuple
        """
        assert rdkit_ff_type.lower() in ("mmff", "uff")
        if refine_with_crest:
            from tempfile import TemporaryDirectory
            from subprocess import run

            # We need both CREST and xTB installed.
            if run("crest --version", shell=True).returncode != 0:
                raise RuntimeError(
                    "`CREST` is not found! Make sure it is installed and added into the PATH."
                )
            if run("xtb --version", shell=True).returncode != 0:
                raise RuntimeError(
                    "`xTB` is not found! Make sure it is installed and added into the PATH."
                )
        mol = MolFromSmiles(smiles)
        mol = AddHs(mol)
        AllChem.EmbedMultipleConfs(mol, numConfs=num_conformers, params=AllChem.ETKDG())
        symbols = [atom.GetSymbol() for atom in mol.GetAtoms()]
        energies = []
        for conf_id in range(num_conformers):
            if rdkit_ff_type.lower() == "mmff":
                ff = AllChem.MMFFGetMoleculeForceField(
                    mol, AllChem.MMFFGetMoleculeProperties(mol), confId=conf_id
                )
            else:  # UFF
                ff = AllChem.UFFGetMoleculeForceField(mol, confId=conf_id)
            energy = ff.CalcEnergy()
            energies.append((conf_id, energy))
        lowest_energy_conf = min(energies, key=lambda x: x[1])
        coordinates = mol.GetConformer(id=lowest_energy_conf[0]).GetPositions()
        if refine_with_crest:
            xyz = f"{len(symbols)}\n\n" + "\n".join(
                f"{s} {coordinates[i][0]:.10f} {coordinates[i][1]:.10f} {coordinates[i][2]:.10f}"
                for i, s in enumerate(symbols)
            )
            chrg = GetFormalCharge(mol)
            uhf = int(spin * 2)
            with TemporaryDirectory(dir=Path.cwd()) as temp_dir:
                with open(Path(temp_dir) / "mol.xyz", "w", encoding="utf-8") as f:
                    f.write(xyz)
                s = run(
                    f"crest mol.xyz -gfn2 -quick -prop ohess{f' --chrg {chrg}' if chrg != 0 else ''}{f' --uhf {uhf}' if uhf != 0 else ''}",
                    shell=True,
                    cwd=temp_dir,
                )
                if s.returncode == 0:
                    with open(Path(temp_dir) / "crest_property.xyz", "r") as f:
                        xyz = f.readlines()
                    xyz_data = []
                    for i in xyz[2:]:
                        if i == xyz[0]:
                            break
                        xyz_data.append(i.strip().split())
                    xyz_data = np.array(xyz_data)
                    symbols, coordinates = np.split(xyz_data, [1], axis=-1)
                    symbols = symbols.flatten().tolist()
                    coordinates = coordinates.astype(np.float64)
        return symbols, coordinates

    def cartesian2smiles(
        self,
        symbols: List[str],
        coordinates: np.ndarray,
        charge: int = 0,
        canonical: bool = True,
    ) -> str:
        """
        Transform (guess out) molecular geometry to SMILES string.

        :param symbols: a list of atomic symbols
        :param coordinates: Cartesian coordinates;  shape: (n_a, 3)
        :param charge: net charge
        :param canonical: whether to canonicalise the SMILES
        :type symbols: list
        :type coordinates: numpy.ndarray
        :type charge: int
        :type canonical: bool
        :return: SMILES string
        :rtype: str
        """
        mol = self._xyz2mol(symbols, coordinates)
        rdDetermineBonds.DetermineBonds(mol, charge=charge)
        smiles = MolToSmiles(mol)
        if canonical:
            smiles = CanonSmiles(smiles)
        return smiles
